/*
** Module   : TerminalMenu.java
** Abstract : menu implementation for the terminal simulator
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20150420 Support the content pane as a separate object from the ChUI simulator.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import javax.swing.*;

import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * Provides the menu implementation for the terminal simulator.
 */
public class TerminalMenu
extends JPopupMenu
{
   /** Select Text checkbox menu item. */
   private JCheckBoxMenuItem select = new JCheckBoxMenuItem("Select Text");
   
   /** Copy menu item. */
   private JMenuItem copy = new JMenuItem("Copy");
   
   /** Paste menu item. */
   private JMenuItem paste = new JMenuItem("Paste");
   
   /** Screen Capture to Clipboard menu item. */
   private JMenuItem capture = new JMenuItem("Screen Capture to Clipboard");
   
   /** Print menu item. */
   private JMenuItem print = new JMenuItem("Print");
   
   /** Settings menu item. */
   private JMenuItem settings = new JMenuItem("Settings");
   
   /** Quit menu item. */
   private JMenuItem exit = new JMenuItem("Exit");
   
   /** Parent frame. */
   private Frame frame = null;
   
   /** CHUI simulator instance. */
   private ChuiSimulator sim = null;
   
   /** Content  pane on top of which we are processing. */
   private ContentPane pane = null;
   
   /** Configuration storage. */
   private TerminalOptions options = null;
   
   /** System clipboard instance. */
   private Clipboard clipboard = null;
   
   /**
    * Create a menu instance, add menu items and setup event listeners.
    *
    * @param    frame
    *           The parent frame.
    * @param    sim
    *           The CHUI simulator instance.
    * @param    pane
    *           The content pane on top of which we are processing.
    * @param    container
    *           The outermost container which should be resized on settings
    *           changes. This also determines if an "Exit" menu item should
    *           be exposed.
    * @param    options
    *           The terminal configuration.
    */
   public TerminalMenu(final Frame           frame,
                       final ChuiSimulator   sim,
                       final ContentPane     pane,
                       final ContainerOps    container,
                       final TerminalOptions options)
   {
      super();
      
      // add items
      add(select);
      add(copy);
      add(paste);
      add(capture);
      add(print);
      add(settings);
      
      // only applications can exit
      if (container.canExit())
      {
         add(exit);
      }
      else
      {
         exit = null;
      }
      
      // by default these items are disabled, events will later determine
      // if they can be enabled
      copy.setEnabled(false);
      paste.setEnabled(false);
      
      this.frame   = frame;
      this.sim     = sim;
      this.pane    = pane;
      this.options = options;
      
      try
      {
         clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
         enablePaste();
         
         clipboard.addFlavorListener(new FlavorListener()
         {
            public void flavorsChanged(FlavorEvent evt)
            {
               enablePaste();
            }
         });
      }
      
      catch (Exception exc)
      {
         // either we are in headless mode or there is a security exception
         // so we ignore the error and copy/paste will be disabled since
         // the clipboard instance is null
      }
      
      select.addItemListener(new ItemListener()
      {
         public void itemStateChanged(ItemEvent evt) 
         {
            int change = evt.getStateChange();
            
            switch (change)
            {
               case ItemEvent.SELECTED:
                  sim.enableSelectionTrack(true);
                  break;
               case ItemEvent.DESELECTED:
                  sim.cancelSelection();
                  break;
            }
         }
      });
      
      copy.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            String data = sim.copy();
            
            if (data != null)
            {
               clipboard.setContents(new StringSelection(data), null);
            }
         }
      });
      
      paste.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            Transferable trans = clipboard.getContents(pane);
            
            if (trans != null)
            {
               String data = null;
               
               try
               {
                  data = (String) trans.getTransferData(DataFlavor.stringFlavor);
               }
               
               catch (Exception exc)
               {
                  // may have UnsupportedFlavorException or IOException but
                  // this just means we can't paste, so ignore
               }
                  
               if (data != null)
               {
                  pane.paste(data);
                  
                  // clear the current selection
                  sim.cancelSelection();
               }
            }
         }
      });
      
      capture.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            if (clipboard != null)
            {
               Image image = pane.capture();
               
               if (image != null)
               {
                  clipboard.setContents(new ImageXfer(image), null);
               }
            }
         }
      });
      
      print.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            pane.print();
         }
      });
      
      settings.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            // this MUST be called using the containing class' member since
            // otherwise we would always reference the first options instance
            // (since the TerminalMenu constructor is only called once but the
            // action may be performed many times); since each call should take
            // the most recent settings, this must be cached
            TerminalOptions ed = SettingsPanel.edit(frame, TerminalMenu.this.options);

            if (!ed.equals(options))
            {
               // remember our edits for later 
               TerminalMenu.this.options = ed;
               
               sim.configure(ed);
               container.forceResize();
            }
            
            sim.cancelSelection();
         }
      });
      
      if (container.canExit())
      {
         exit.addActionListener(new ActionListener()
         {
            public void actionPerformed(ActionEvent evt) 
            {
               pane.quit();
               System.exit(0);
            }
         });
      }
   }
   
   /**
    * Notify the select text menu item that the selection operation is
    * complete.  This will cause an item changed deselected event which
    * will call the simulator to clear its state.
    */
   void textSelectionDone()
   {
      select.setSelected(false);
   }
   
   /**
    * Enable or disable the copy menu item.
    *
    * @param    enable
    *           <code>true</code> to enable the menu item.
    */
   void enableCopy(boolean enable)
   {
      if (clipboard != null)
      {
         copy.setEnabled(enable);
      }
   }
   
   /**
    * Enable or disable the paste menu item based on whether the clipboard
    * does or does not (respectively) currently have string data that can be
    * pasted.
    */
   void enablePaste()
   {
      if (clipboard != null)
      {
         paste.setEnabled(clipboard.isDataFlavorAvailable(DataFlavor.stringFlavor));
      }
   }
}

