/*
** Module   : ChuiWebSimulator.java
** Abstract : a WebSocket-based ChUI terminal simulator
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 MAG 20140211 Implements COLOR.
** 005 CA  20140206 Web socket timeout can't be read from the jetty threads (as these don't have a
**                  P2J context, so it must be passed at construction.
** 006 MAG 20140217 Refactoring web client redirection flaw.
** 007 MAG 20140219 Improve performances.
** 008 MAG 20140402 Implements interactive process.
** 009 MAG 20140702 Implements copy/paste.
** 010 CA  20140805 Identify this simulator as a web simulator.
** 011 MAG 20140821 Extends color palette.
** 012 CA  20141206 Fixed a starvation problem - wait for the next key to be read, before 
**                  returning from readKey().
** 013 GES 20150318 Refactored to move websockets code into a separate class for sharing.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.cfg.*;

/**
 * This is a WebSocket-based ChUI terminal simulator which uses a binary protocol to
 * communicate with the javascript (browser-based) terminal simulator.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application is blocked.
 */
public class ChuiWebSimulator
extends EmulatedTerminalState
implements WebClientMessageTypes,
           ClientProtocolHooks
{  
   /** Color palette message spec. */
   private static final String COLOR_PALETTE = "{\"c\":0, \"p\":%s}"; 
   
   /** Internal locking object which is also shared with the websock instance. */
   private final Object lock = new Object();
     
   /** Color palette JSON message. */
   private String paletteMessage;
   
   /** Screen contents. */
   private Cell[][] cache = null;
   
   /** Worker object to processing our WebSocket that connects us to the websock client. */
   private WebClientProtocol websock = null;
   
   /** VT100 keyboard instance (only used for VT100 emulation in child processes). */
   private VT100ProgressKeyboard vt100Keyboard =  null;
   
   /**
    * Create a new chui web simulator.
    * 
    * @param    config
    *           Web socket timeout.
    */
   public ChuiWebSimulator(BootstrapConfig config)
   {
      this.paletteMessage = buildPaletteMessage();
      
      long timeout = -1;
      long wdtimeout = -1;
      
      if (config != null)
      {
         timeout   = config.getInt("client", "web", "socketTimeout", -1);
         wdtimeout = config.getInt("client", "web", "watchdogTimeout", -1);
      }
      
      websock = new WebClientProtocol(lock, this, timeout, wdtimeout);
   }
   
   /**
    * Obtain the web client protocol instance that implements the web socket operations.
    *
    * @return   The web socket protocol implementation.
    */
   public WebClientProtocol getWebSocket()
   {
      return websock;
   }
   
   /**
    * Called each time a connect occurs, this allows caller-controlled initialization of the
    * remote client.  There can only be 1 connection open at a time, but it is possible for
    * the connection to be opened, closed and then re-opened.
    * <p>
    * Whenever the main page is loaded or re-loaded, the WebSocket will connect and we must
    * put the virtual terminal into a known state.
    */
   @Override
   public void initRemoteClient()
   {
      clearCache();
      sendColorPalette();
      
      // synchronize
      websock.sendSwitchMode();
      triggerRepaint(true);
      setCursorStatus(showCursor);
      
      if (showCursor)
      {
         setCursor(getCursorX(), getCursorY());
      }
   }
   
   /**
    * Input key events when in VT100 mode.
    * 
    * @param    keyCode
    *           The key code from the input event.
    * @param    charCode
    *           The character code from the input event.
    */
   @Override
   public void injectVT100Key(int keyCode, char charCode)
   {
      vt100Keyboard.handle(keyCode, charCode);
   }
   
   /**
    * Assign the VT100 emulator keyboard instance.
    *
    * @param    vt100Keyboard
    *           The keyboard instance.
    */
   public void setVT100Keyboard(VT100ProgressKeyboard vt100Keyboard)
   {
      this.vt100Keyboard = vt100Keyboard;
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      return websock.readKey();
   }
   
   /**
    * Initialize the instance using the given configuration.
    * Allocate screen cache. 
    *
    * @param   options
    *          The configuration for the terminal simulator.
    */   
   @Override
   public void configure(TerminalOptions options)
   {
      super.configure(options);
      
      synchronized (lock)
      {
         if (cache == null)
         {
            clearCache();
         }
      }
   }

   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0. Send INIT message down to JS client.
    */   
   @Override
   public void clear()
   {
      super.clear();
      websock.sendBinaryMessage(MSG_CLEAR);
   }

   /**
    * Raise the bell on the remote terminal.
    */
   public void beep()
   {
      websock.beep();         
   }
   
   /**
    * Set the cursor visibility. Send SET_CURSOR_ON or SET_CURSOR_OFF message down to JS client.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visibility state.
    */   
   @Override
   public boolean setCursorStatus(boolean on) 
   throws IllegalStateException
   {
      websock.sendBinaryMessage(MSG_CURSOR_STATUS, (byte)(on ? 1 : 0));
      
      // call super implementation
      return super.setCursorStatus(on);
   }

   /**
    * Position the cursor. Send SET_CURSOR_POS message down to JS client.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean setCursor(int x, int y)
   {
      // TODO: this will limit our maximum cursor position to 127 which may not be enough
      //       in some configurations!
      websock.sendBinaryMessage(MSG_CURSOR_POS, (byte) y, (byte) x);
      
      // call super implementation
      return super.setCursor(x, y);
   }
   
   /**
    * When data is prepared to send down to the websock JS code to be drawn, the data should be 
    * optimized for drawing on a canvas.
    * <p>
    * The video screen memory is composed of cells stored in a 2 dimensional matrix.  The size
    * of the matrix is rows x columns where rows are the number of rows in the screen and
    * columns are the number of columns. For example, 24 x 80 for a screen having 24 rows 
    * and 80 columns. Each cell is a structure which holds cell attributes, color and character.
    * <p>
    * The state of the most recently updated screen is saved in a cache. Whenever a new screen  
    * should be sent to the websock, the new screen is compared with the cached screen and only
    * the changed cells are sent.  Then the cache is updated to reflect the latest changes. A 
    * cell is considered changed when at least one of the cell properties (attribute, color or 
    * the character) is different from the same cell in the cache.
    * <p>
    * Drawing text on the HTML5 canvas using JS code is done on two steps:
    * <ol> 
    *  <li> The background rectangle containing the text is filled with the background color. 
    *       Maximizing the size of the filled rectangle optimizes the drawing.
    *  <li> The second step is to fill the rectangle with the contained text using the foreground 
    *       color. On HTML5 canvas this can be done only on a line basis by filling with text 
    *       rectangles having the height equal to one character. An optimized draw on this case 
    *       requires the text length to be as long as possible.
    * </ol>
    * <p>
    * When a network message is constructed this is done on two steps by keeping in mind the 
    * performance constraints described above.
    * <ul>
    *  <li> A new screen to be sent is scanned a line at a time, column by column to detect 
    *       changed cells. Adjacent changed cells on a line having the same color and attributes 
    *       are stored together in a <code>Chunk</code> structure. The chunks are stored in a
    *       linked map structure having the cell Color (attribute and color) structure as a key
    *       and a linked list of chunks as the value. Both the linked map and the linked list   
    *       preserve the insertion order.
    *  <li> On the next step the message is built and contains all chunks from the screen as 
    *       detected in the first step. For each entry in the map the list of chunks is used to
    *       build rectangle structures. The chunks are extracted from the list one by one. If
    *       multiple chunks could be combined to form larger rectangles, the chunks are combined
    *       to form the largest possible rectangle. The width of a rectangle is the number of 
    *       characters from a chunk and the height are the number of chunks combined. For example
    *       a rectangle of 25 x 5 holds 5 lines of 25 text characters each. All characters have  
    *       in a chunk have the same color and text attributes. The condition for two chunks to
    *       form a rectangle is to have the same length, to start at the same position within the
    *       line and to be one adjacent lines. Chunks that cannot be combined as large rectangles
    *       are stored inside the message as simple rectangles, each having a one character
    *       height.
    * </ul>
    * 
    * @param    full
    *           Flag to force a the entire screen to repaint.
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      int rows    = getRows();
      int columns = getColumns();
      
      synchronized (lock)
      {
         /* chunks map */
         Map<Color, List<Chunk>> chunks = new LinkedHashMap<>();
         
         // chunk
         ByteArrayOutputStream chunk = null;
         
         // get screen data
         Cell[][] data = getScreenData();
               
         for (int row = 0; row < rows; row++)
         {             
            // attributes of previous cell
            Color previous = null;

            for (int col = 0; col < columns; col++)
            {
               Cell cell = data[row][col];
               
               // check if cell has changes
               if (cell != null && !cell.equals(cache[row][col]))
               {
                  Color color = cell.color(); 

                  // check for color and attributes changes
                  if (color != null && color.equals(previous))
                  {
                     // same color and attributes.
                     // add cell to the current chunk.
                     putCellCharacter(chunk, cell);
                  }
                  else
                  {
                     // attributes changed.
                     // add current chunk if exists and start a new one.
                     addChunk(chunks, chunk, previous);
                     // new chunk
                     previous = color;
                     chunk = new ByteArrayOutputStream();
                     chunk.write(row); // row number                     
                     chunk.write(col); // column number
                     putCellCharacter(chunk, cell);
                  }
               }
               else
               {
                  // the cell is already in cache do not send. 
                  // add current chunk and reinitialize. 
                  addChunk(chunks, chunk, previous);
                  previous = null;
               }
               
               cache[row][col] = cell;               
            }
            // start a new line
            // check if a chunk exists and add to the list of chunks.
            addChunk(chunks, chunk, previous);
         }   
         
         // build the message from chunks map.
         if (!chunks.isEmpty())
         {
            websock.sendBinaryMessage(buildDrawMessage(chunks, rows));
         }
      }
   }
   
   /**
    * Remote application is about to end. Notify the javascript client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      websock.quit();         
   }

   /**
    * Build color palette message. This message is the same
    * over the user session so it's build and store on constructor.
    * The message is send whenever a WebSocket connection is open.
    *
    * @return  A string representing the JSON format of color palette.
    */
   private String buildPaletteMessage()
   {
      StringBuilder msg = new StringBuilder("[");
      
      for (ColorPalette clr : ColorPalette.values())
      {
         if (msg.length() != 1)
         {
            msg.append(",");
         }
         
         String rgb = String.format("#%02x%02x%02x",
                                    clr.getRed(),
                                    clr.getGreen(),
                                    clr.getBlue());
         
         msg.append("\"").append(rgb).append("\"");
      }      
      msg.append("]");
      
      // JSON format
      return String.format(COLOR_PALETTE, msg.toString());
   }
   
   /**
    * Send the color palette down to the javascript client. This should be done whenever the
    * WebSocket connection is opened.
    */
   private void sendColorPalette()
   {      
      websock.sendTextMessage(paletteMessage);
   }
   
   /**
    * Add a chunk to the chunks map.
    * 
    * @param    chunks
    *           Chunks map.
    * @param    chunk
    *           Chunk to add.
    * @param    color
    *           Color attribute used as a key
    */
   private void addChunk(Map<Color, List<Chunk>> chunks, 
                         ByteArrayOutputStream   chunk,
                         Color                   color)
   {
      if (color != null && chunk != null && chunk.size() > 0)
      {
         List<Chunk> values = chunks.get(color);
         
         if (values == null)
         {
            values = new LinkedList<Chunk>();
            chunks.put(color, values);
         }
         
         values.add(new Chunk(chunk.toByteArray()));
      }
   }
   
   /**
    * Build the screen drawing message from a given chunks map.
    *  
    * @param    chunks
    *           Chunks map.
    * @param    rows
    *           The number of rows in the screen.
    *
    * @return   The completed binary message.
    */
   private byte[] buildDrawMessage(Map<Color, List<Chunk>> chunks, int rows)
   {
      // message
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_DRAW);
      
      for (Map.Entry<Color, List<Chunk>> entr : chunks.entrySet())
      {
         Color color = entr.getKey();
         // list of chunks having the same color
         LinkedList<Chunk> list = (LinkedList<Chunk>) entr.getValue();
         // build a message as rectangles
         while (!list.isEmpty())
         {
            // rectangle height
            int height = 1;
            // poll first chunk
            Chunk chunk = list.pollFirst();
            // try to build a rectangle
            ByteArrayOutputStream rect = new ByteArrayOutputStream();
            rect.write(0); // packet size
            rect.write(0);
            rect.write(chunk.getTextSize() / 2);  // width
            rect.write(1);  // height
            rect.write(chunk.getRow());  // row
            rect.write(chunk.getColumn());  // col
            // add attributes and color
            putCellAttributes(rect, color);
            // write text
            rect.write(chunk.getChunk(), 
                       Chunk.HEADER_SIZE, 
                       chunk.getTextSize());
            // check on the next line
            for (int i = chunk.getRow() + 1; i < rows; i++)
            {
               boolean rectangleDetected = false;
               
               Chunk next = new Chunk(i, chunk.getColumn(), chunk.getSize());
               
               // search a candidate to add to rectangle
               for (Chunk candidate : list)
               {
                  // detect rectangle 
                  rectangleDetected = candidate.equals(next); 

                  // if found rectangle
                  if (rectangleDetected)
                  {
                     height++;
                     // consolidate rectangle
                     rect.write(candidate.getChunk(), 
                                Chunk.HEADER_SIZE, 
                                candidate.getTextSize());
                     list.remove(candidate);
                     break;
                  }               
               }               
               
               // end of rectangle detection. 
               if (!rectangleDetected)
               {
                  break;
               }
            }
            // update packet size, rectangle height and add to message.
            byte[] rectBytes = rect.toByteArray();
            int packetSize = rectBytes.length;
            rectBytes[0] = (byte)((packetSize >>> 8) & 0xFF);
            rectBytes[1] = (byte) (packetSize & 0xFF);
            rectBytes[3] = (byte) height;
            message.write(rectBytes, 0, rectBytes.length);
         }
      }
      
      return message.toByteArray();
   }
   
   /**
    * Switch mode message.
    */
   private void sendSwitchMode()
   {
      websock.sendSwitchMode();               
   }
   
   /**
    * Put cell attributes and character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cellColor
    *          Cell color object.
    */
   private void putCellAttributes(ByteArrayOutputStream message, Color cellColor)
   {
      // attributes
      if (cellColor == null)
      {
         // attributes
         message.write(0);
         // color
         message.write(0);                        
      }
      else
      {
         // attributes
         message.write(cellColor.attribute());
         // color
         int color = cellColor.color();
         if (color == 0)
         {
            message.write(0);               
         }
         else
         {
            int bg = color & 0x0F;
            int fg = (color >>> 12) & 0xF0;
            message.write(fg | bg);
         }               
      }
   }
   
   /**
    * Put cell character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cell
    *          Cell object.
    */
   private void putCellCharacter(ByteArrayOutputStream message, Cell cell)
   {
      // character
      char c = cell.character();
      message.write((c >>> 8) & 0xFF);
      message.write(c & 0xFF);
   }
   
   /**
    * Clear the cached screen.
    */
   private void clearCache()
   {
      cache = new Cell[getRows()][getColumns()];
   }
}