/*
** Module   : LinuxKeyboardReader.java
** Abstract : process swing key events and convert them to 4GL keys, compatible with Linux clients.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141029 Created initial version.
** 002 GES 20150506 Moved to a driver package.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * This class represents a {@link SwingKeyboardReader} implementation for clients which require
 * NCURSES/Linux compatibility.
 * 
 * Known Issues and/or TODOs
 * <p>
 * <ul>
 *   <li> there are still MANY key code differences that can occur in the 
 *        CTRL+<key>, ALT+SHIFT+<key> cases (and many others), these differences
 *        are generally only seen in the cases of keys not processed via
 *        keyTyped(); the general behavior in these cases is for the wrong key
 *        code or no key code to be generated; it is unknown if these cases
 *        are really useful but NCURSES does generate them
 *   <li> all key codes are being generated to be compatible with the NCURSES
 *        CHUI implementation, which is known to be at least slightly different
 *        from the 4GL implementation 
 * </ul> 
 */
public class LinuxKeyboardReader
extends SwingKeyboardReader
{
   /** Track events that occur when the ESC key is pressed. */
   private boolean escPressed = false;
   
   /** Translation map of Progress key codes to Progress alt key codes. */
   private final Map<Integer, Integer> xlateAlt = new HashMap<>();

   /**
    * Adds special translations related to NCURSES.
    */
   public LinuxKeyboardReader()
   {
      // ALT+<key> translation
      
      // The following are not handled here:
      //    VK_BACKSPACE (handled in keyTyped())
      //    VK_BACKTAB (intercepted by OS, no info on what the code shouuld be)
      //    VK_CTRL_*  (handled in keyTyped())
      //    VK_DELETE (custom logic in keyTyped())
      //    VK_ESCAPE (this is never returned to Progress code anyway)
      //    VK_TAB (intercepted by OS, no info on what the code shouuld be)
      
      // ncurses and other terminal processing treats the ALT key as the same
      // as having previously pressed ESC in some cases; in the cases in the
      // following table, a different code is generated but it is NOT compatible
      // with the ESC processing; this is why this add'l table exists
      
      // WARNING: the hard coded values have no known meaning but were found
      //          to be generated by the P2J CHUI NCURSES implementation
      
      xlateAlt.put(Key.VK_CURSOR_DOWN, 518);
      xlateAlt.put(Key.VK_END, 523);
      xlateAlt.put(Key.VK_ENTER, Key.VK_ENTER | 1024);
      xlateAlt.put(Key.VK_F1, Key.VK_F13);
      xlateAlt.put(Key.VK_F2, Key.VK_F14);
      xlateAlt.put(Key.VK_F3, Key.VK_F15);
      xlateAlt.put(Key.VK_F4, Key.VK_F16);
      xlateAlt.put(Key.VK_F5, Key.VK_F17);
      xlateAlt.put(Key.VK_F6, Key.VK_F18);
      xlateAlt.put(Key.VK_F7, Key.VK_F19);
      xlateAlt.put(Key.VK_F8, Key.VK_F20);
      xlateAlt.put(Key.VK_F9, Key.VK_F21);
      xlateAlt.put(Key.VK_F10, Key.VK_F22);
      xlateAlt.put(Key.VK_F11, Key.VK_F23);
      xlateAlt.put(Key.VK_F12, Key.VK_F24);
      xlateAlt.put(Key.VK_HOME, 528);
      xlateAlt.put(Key.VK_INSERT, 533);
      xlateAlt.put(Key.VK_CURSOR_LEFT, 538);
      xlateAlt.put(Key.VK_PAGE_DOWN, 543);
      xlateAlt.put(Key.VK_PAGE_UP, 548);
      xlateAlt.put(Key.VK_CURSOR_RIGHT, 553);
      xlateAlt.put(Key.VK_CURSOR_UP, 559);
   }
   
   /**
    * This method is called when any key is pressed or when a multi-key
    * combination (CTRL, SHIFT, ALT modifiers) is pressed.  Key events
    * provided here are low level events and have not been processed into
    * a higher level code as is done in {@link #keyTyped}.
    * <p>
    * Once a valid key code is generated, that key code is added to the key
    * input queue in FIFO order. Any keys that are not recognized as having
    * valid counterparts in the 4GL key code mappings, are simply ignored
    * (nothing is added to the key input queue for unrecognized keys).
    * <p>
    * This method is used to generate key codes 128 and above. There are 3
    * paths through this method:
    * <p>
    * <ol>
    *   <li> If the key is VK_ESCAPE, a flag for tracking a pressed escape is
    *        toggled and a key of -1 is added to the queue.
    *   <li> A map lookup of the Java key code (as provided in the event) is
    *        executed. If there is a matching 4GL key code, then a key is most
    *        likely going to be inserted in the queue. See below for more
    *        details.
    *   <li> Otherwise, the key press is ignored.
    * </ol>
    * <p>
    * When the map lookup of the 4GL key code occurs, usually the returned 4GL
    * key code is posted into the end of the key input queue without change.
    * But there are some cases in which the 4GL key code is modified first.
    * <p>
    * If the 4GL key code is VK_TAB and the SHIFT modifier key is on (the
    * shift key was pressed at the time the tab was pressed), then a VK_BACKTAB
    * is posted to the queue.  Otherwise, the VK_TAB is ignored because it is
    * naturally processed in <code>keyTyped()</code> (but the back tab is not).
    * <p>
    * If the escape key was not pressed, but the ALT modifier was pressed when
    * this key code was generated, then a 2nd level of translation table
    * lookups occurs. The 4GL key code is checked in an ALT translation table.
    * If there is a mapped value, that value replaces the previously generated
    * 4GL key code and is placed in the queue.
    * <p>
    * All other ESC and ALT processing is the same as already described in the
    * <code>keyTyped</code> method.
    *
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyPressed(KeyEvent evt)
   {
      int     code  = -1;
      int     key   = evt.getKeyCode();
      boolean shift = evt.isShiftDown();
      boolean alt   = evt.isAltDown();
      
      // escape is never returned as 27, instead it sets some state as a
      // modifier key
      if (key == KeyEvent.VK_ESCAPE)
      {
         // toggle the flag
         escPressed = !escPressed;
         
         queue.add(-1);
         return;
      }
      
      Integer mappedCode = xlate.get(key);
      
      if (mappedCode != null)
      {
         code = mappedCode.intValue();
         
         // normal VK_TAB is generated naturally in keyTyped() but SHIFT +
         // TAB must be handled here
         if (code == Key.VK_TAB)
         {
            code = shift ? Key.VK_BACK_TAB : -1;
         }
         
         if (code != -1)
         {
            if (alt)
            {
               if (!escPressed)
               {
                  // process special ALT+<key> combinations (but only if the
                  // escape flag was OFF, since otherwise
                  Integer altCode = xlateAlt.get(code);
                  
                  if (altCode != null)
                  {
                     code = altCode.intValue();
                  }
               }
               else
               {
                  // disable escape flag and bail
                  escPressed = false;
                  
                  if (code != Key.VK_ENTER)
                     code = -1;
               }
            }
            else
            {
               // honor the escape modifier if it was on previously
               code = processEscape(code, false);
               
               // if escape processing disabled the key, exit now
               if (code == -1)
                  return;
            }
         }
         
         if (code != -1)
         {
            queue.add(code);
         }
      }
      
      /*
      if (code == -1)
      {
         System.out.printf("%s\n", evt.toString());
      }
      */
   }

   /**              
    * This method is called when a pressed key (or a multi-key combination) is
    * converted to a higher level representation. As a general rule, all
    * key codes between 0 and 127 inclusive are generated through this
    * callback. The reason for this is that Java translates the keys or
    * multi-key combinations into the higher level representation which almost
    * perfectly matches the 4GL key codes 0 through 127.  In other words, the
    * keyboard inputs of the user are very compatible between the 4GL and Java
    * for those portions of the key codes (0 through 127).
    * <p>
    * Once a valid key code is generated, that key code is added to the key
    * input queue in FIFO order.
    * <p>
    * The non-visual values 0 through 31 are handled almost "for free". The
    * user generates these in the following manner:
    * <p>
    * <pre>
    * Key Code  User Input               Notes
    * --------  -----------------------  --------------------------------------
    * 0         CTRL+2                   Special code in this method supports this.
    * 1         CTRL+A
    * 2         CTRL+B
    * 3         CTRL+C
    * 4         CTRL+D
    * 5         CTRL+E
    * 6         CTRL+F
    * 7         CTRL+G
    * 8         CTRL+H or VK_BACKSPACE
    * 9         CTRL+I or VK_TAB         The containing frame's focus traversal
    *                                    keys must be forced to the empty set in
    *                                    order for this to work.
    * 10        CTRL+J or VK_ENTER       NOT DELIVERED, handled in keyPressed()
    * 11        CTRL+K
    * 12        CTRL+L
    * 13        CTRL+M
    * 14        CTRL+N
    * 15        CTRL+O
    * 16        CTRL+P
    * 17        CTRL+Q
    * 18        CTRL+R
    * 19        CTRL+S
    * 20        CTRL+T
    * 21        CTRL+U
    * 22        CTRL+V
    * 23        CTRL+W
    * 24        CTRL+X
    * 25        CTRL+Y
    * 26        CTRL+Z
    * 27        VK_ESCAPE                NOT DELIVERED, 4GL doesn't ever return this.
    * 28        CTRL+\
    * 29        CTRL+]
    * 30        CTRL+6                   Special code in this method supports this.
    * 31        CTRL+-                   Special code in this method supports this.
    * </pre>
    * <p>
    * Two of the above codes are never returned in Progress (10 and 27) and so
    * they are excluded here too.
    * <p>
    * Key code 10 is the ASCII linefeed and it is generated from either the
    * CTRL+J or the VK_ENTER key combinations in Java. These combinations
    * always generate a key code of 13 in Progress, which is the ASCII carriage
    * return code. Since this is a translation of Java codes to 4GL key codes,
    * this method bypasses these key combinations and logic exists in
    * keyPressed() to handle the translation.
    * <p>
    * The key code 27 is the ASCII escape code. In character terminals it is
    * used to notify the application of a special condition. Progress did this
    * "eating" the escape key and then usually encoding the next character as
    * its normal key code bitwise OR'd with 1024. In other words, they encode
    * the fact that there was a preceding escape key my encoding a modifier
    * in a more significant bit that cannot be set directly by user keyboard
    * input. It is important to note that the escape key is pressed, then
    * released and then the next key that is pressed is modified and sent.
    * This means that pressing the escape key causes state to be stored so that
    * the next key can be modified.
    * <p>
    * Terminals (and thus the 4GL which was originally designed to work
    * exclusively as a terminal application) give the ALT key an ESC-like
    * behavior.  When used instead of pressing a preceding ESC key, it generates
    * the same modified result. The primary difference is that it must be
    * pressed at the same time as the key to be modified.
    * <p>
    * For example, pressing the '2' key generates a key code of 50 which is the
    * ASCII code for the digit 2. Pressing ESC (and releasing it) and then
    * pressing the '2' key, generates the key code 1074 (which is 50 | 1024).
    * The key code 1074 is also generated when the user holds the ALT key down
    * while simultaneously pressing the '2' key.
    * <p>
    * It should be noted that pressing ESC and then ESC will clear the special
    * mode. In other words, the next character sent to the server after 2
    * sequential escape keys will be the unmodified key code, just as if the
    * escape key was never pressed in the first place.
    * <p>
    * The ALT key has a similar affect. If an ESC key is pressed and then an
    * ALT+key combination is pressed, the result is the unmodified key code.
    * The ALT key negates the previous ESC state. So the modification of key
    * codes only occurs if a single escape was the most recent key AND ALT is
    * not pressed OR the last key was not an escape key and the ALT key was
    * pressed.
    * <p>
    * To be clear, the ALT key cannot be pressed and release on its own. In
    * such a case it doesn't have the effect of escape, but is instead just
    * ignored.  The ALT key MUST be held down while another key is pressed
    * for it to have the modification affect.
    * <p>
    * When there is any use of ESC/ALT with the lowercase alphabetic characters
    * 'a' through 'z', the code uppercases them before modification. The reasons
    * for doing this are not clear, but there was this comment in the original
    * NCURSES-based key processing implementation:
    * <p>
    * <pre>
    *    Remapping lowercase Latin to uppercase. This differs
    *    with Progress but seems to be more convenient to user because
    *    Ctrl+<letter> combinations are case insensitive (and even
    *    keyboard layout insensitive). Note the usage of direct
    *    calculation not Character class methods. This is
    *    because we are dealing with integer key codes not characters
    *    so this is dirty hack anyway.
    * </pre>
    * <p>
    * Finally, the escape driven modification of key codes is only done if the
    * non-modified key code is found to have a non-empty string as a key label.
    * This is determined by a call to <code>Keyboard.keyLabel(code)</code> and
    * checking the returned string.
    * <p>
    * Key codes 0, 30 and 31 are all created based on CTRL + a key that is in
    * the number row of the keyboard.  Normally, CTRL+@ generates 0, CTRL+^
    * generates 30 and CTRL+_ generates 31. In Java, @, ^ and _ characters can
    * only be generated using the SHIFT key and the '2', '6' or '-' keys
    * respectively.  In Progress, no shift key is needed to generate the
    * resulting 0, 30 or 31 key codes.  While all the other CTRL+key codes
    * between 0 and 31 are handled by Java naturally in the keyTyped() method,
    * the proper support for 0, 30 and 31 is implemented by custom logic in this
    * method.
    * <p>                               
    * Key codes between 32 and 127 inclusive are naturally created by Java
    * and are passed to this method ready to go.  This includes the non-visual
    * VK_DELETE key which is mapped to 127.  All other characters are visual
    * and their key codes are naturally created by Java by their common key
    * on the keyboard OR by their common key + SHIFT.  For example, 'g'
    * generates 103 (the ASCII code for 'g') and SHIFT+'g' generates 71 which
    * is the ASCII code for 'G'. '4' generates code 52 (ASCII for the digit 4)
    * and SHIFT+'4' (on a U.S. keyboard layout) generates 36 which is the
    * ASCII code for '$'.  This is all exactly the same as how it works in
    * the NCURSES-based implementation.  The only exception to the above is
    * that there is custom logic to generate the key code 512 when ESC then
    * VK_DELETE (or ALT+VK_DELETE) is encountered.  This is an anomaly in
    * how Progress treats the ESC modifier and it had to be dealt with
    * specially.
    *
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyTyped(KeyEvent evt)
   {
      boolean ctrl = evt.isControlDown();
      boolean alt  = evt.isAltDown();
      char    ch   = evt.getKeyChar();
      int     code = -1;
      
      // all CTRL combinations for codes 0 through 31 inclusive are normally
      // provided by the TYPED event, except the following:
      if (ctrl)
      {
         if (ch == '2')
         {
            // CTRL-@
            code = 0;
         }
         else if (ch == '6')
         {
            // CTRL-^
            code = 30;
         }
         else if (ch == '-')
         {
            // CTRL-_
            code = 31;
         }
      }
      
      // normal processing: the preprocessed character is returned directly
      if (code == -1)
      {
         // ESC is never returned directly to Progress code
         // ENTER is processed in keyPressed() and not here
         if (ch != KeyEvent.VK_ESCAPE && ch != KeyEvent.VK_ENTER)
         {
            code = (int) ch;
            
            if ((escPressed && !alt) || (!escPressed && alt))
            {
               // Remapping lowercase Latin to uppercase. This differs
               // with Progress but seems to be more convenient to user because
               // Ctrl+<letter> combinations are case insensitive (and even
               // keyboard layout insensitive). Note the usage of direct
               // calculation not Character class methods. This is
               // because we are dealing with integer key codes not characters
               // so this is dirty hack anyway.
               if (code >= 'a' && code <= 'z')
               {
                  code -= 'a' - 'A';
               }
            }
         }
      }
      
      if (code != -1)
      {
         if (!escPressed && alt && code == Key.VK_DELETE)
         {
            // special case for ALT+DELETE key combo
            code = 512;
         }
         else
         {
            // honor the escape modifier if it was on previously
            code = processEscape(code, alt);
         }
         
         // if escape processing disabled the key, exit now
         if (code == -1)
            return;
         
         queue.add(new Integer(code));
      }
   }
   
   /**
    * If the last key pressed was the escape key, modify the code by bitwise
    * ORing 1024 and return that value if the result has a valid key label as
    * returned by the Progress-compatible <code>Keyboard.keyLabel()</code>. If
    * there is no valid label, -1 will be returned.  If the escape pressed
    * flag is not active, the parameter will be returned unchanged.
    * <p>
    * After this method is called, the escape flag will be off.
    * <p>
    * If the alt modifier is ON but the escape flag is OFF, this method will
    * treat the case as if the escape flag was ON.  If the escape flag is
    * already ON and the alt modifier is ON, then the alt modifier CANCELS the
    * previously pending escape flag and no translation is done.
    *
    * @param    code
    *           The key code for the next key being processed.
    * @param    alt
    *           State of the ALT modifier. <code>true</code> if the modifier
    *           was ON when the given key was pressed.
    *
    * @return   The modified key code if the escape pressed flag is set,
    *           otherwise the input parameter is returned unchanged.
    */
   private int processEscape(int code, boolean alt)
   {
      if ((escPressed && !alt) || (!escPressed && alt))
      {
         // modify the next character (in 4GL, ESC is a modifier key)
         code = code | 1024;
         
         if (Keyboard.keyLabel(code).length() == 0)
         {
            code = -1;
            queue.add(code);
         }
      }
      
      // always toggle the flag off
      escPressed = false;
      
      return code;
   }
}
