/*
** Module   : CaptionButton.java
** Abstract : Define caption button associated with a window.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Changed default BGCOLOR.
** 006 CA  20141021 Fixed drawing of caption buttons; allow restore button to show if window is
**                  maximized.
** 007 MAG 20141029 Added clipping rectangle when drawing.
** 008 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * A caption button associated with a window.
 */
@SuppressWarnings("unchecked")
class CaptionButton
extends Button<GuiOutputManager>
{
   /** Button margin */
   private final static int BUTTON_MARGIN = 3;
   
   /** The button type. */
   private CaptionButtonType type;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The button state - up or down. */
   private boolean up;
   
   /** The button's width, in character units. */
   private double width;
   
   /** The button's height, in character units. */
   private double height;

   /**
    * Create a new button with the specified type.
    * 
    * @param    id
    *           The widget's ID.
    * @param    type
    *           The button type.
    */
   CaptionButton(WidgetId id, CaptionButtonType type)
   {
      super(id, null, "");

      gd = (GuiDriver) OutputManager.getDriver();

      this.type = type;
      this.up = true;
   }

   /**
    * Draw this caption button, depending on its type.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      final WindowGuiImpl window = (WindowGuiImpl) ancestor();
      
      NativeDimension d = physicalDimension();
      final int width = d.width;
      final int height = d.height;

      final NativeRectangle clip = clipRectangle(p, d);
      
      // draw button
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
            gd.fill3DRect(0, 0, width, height, up);
            
            gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
                  
            // draw icon on button
            switch (type)
            {   
               case CLOSE:
                  gd.drawLine(BUTTON_MARGIN,     BUTTON_MARGIN,     width - BUTTON_MARGIN,     height - BUTTON_MARGIN);
                  gd.drawLine(BUTTON_MARGIN + 1, BUTTON_MARGIN,     width - BUTTON_MARGIN,     height - BUTTON_MARGIN - 1);
                  gd.drawLine(BUTTON_MARGIN,     BUTTON_MARGIN + 1, width - BUTTON_MARGIN - 1, height - BUTTON_MARGIN);
                              
                  gd.drawLine(BUTTON_MARGIN,     height - BUTTON_MARGIN,     width - BUTTON_MARGIN,     BUTTON_MARGIN);
                  gd.drawLine(BUTTON_MARGIN,     height - BUTTON_MARGIN - 1, width - BUTTON_MARGIN - 1, BUTTON_MARGIN);
                  gd.drawLine(BUTTON_MARGIN + 1, height - BUTTON_MARGIN,     width - BUTTON_MARGIN,     BUTTON_MARGIN + 1);
                  break;
               case ICONIFY:
                  gd.fillRect(BUTTON_MARGIN, height - BUTTON_MARGIN - 1, width - 2 * BUTTON_MARGIN, 3);
                  break;
               case MAXIMIZE:
                  if (!window.isMaximized())
                  {
                     drawMaxRect(BUTTON_MARGIN, BUTTON_MARGIN, width - 2 * BUTTON_MARGIN, height - 2 * BUTTON_MARGIN);
                  }
                  else
                  {
                     drawMaxRect(BUTTON_MARGIN + width / 4, BUTTON_MARGIN, width / 2, height / 2);
                     
                     gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
                     gd.fillRect(BUTTON_MARGIN, BUTTON_MARGIN + height / 4, width / 2, height / 2);
                     gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));

                     drawMaxRect(BUTTON_MARGIN, BUTTON_MARGIN + height / 4, width / 2, height / 2);
                  }
                  break;
            }
         }
      });
   }

   /**
    * Set the button's dimension.
    * 
    * @param    width
    *           The button's width, in character units
    * @param    height
    *           The button's height, in character units
    */
   public void setDimension(double width, double height)
   {
      this.width = width;
      this.height = height;
   }
   
   /**
    * Get the button's {@link #width}.
    * 
    * @return   See above.
    */
   @Override
   public double width()
   {
      return width;
   }
   
   /**
    * Get the button's {@link #height}.
    * 
    * @return   See above.
    */
   @Override
   public double height()
   {
      return height;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      WindowGuiImpl window = (WindowGuiImpl) ancestor();

      switch (type)
      {   
          case CLOSE:
             // no-op - 4GL doesn't allow closing the app from the close button, 
             // but the WINDOW-CLOSE event is raised
             window.close();
             break;
          case ICONIFY:
             window.iconify();
             break;
          case MAXIMIZE:
             window.maximize();
             break;
      }
      
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      this.up = false;
      repaint();
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      this.up = true;
      repaint();
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Draw a maximum-button style rectangle.
    * 
    * @param    x
    *           The X origin.
    * @param    y
    *           The Y origin.
    * @param    width
    *           The rectangle width.
    * @param    height
    *           The rectangle height.
    */
   private void drawMaxRect(int x, int y, int width, int height)
   {
      gd.drawRect(x, y, width, height);
      gd.drawLine(x, y + 1, x + width, y + 1);
      gd.drawLine(x, y + 2, x + width, y + 2);
   }
}
