/*
** Module   : SwingGuiDriver.java
** Abstract : Swing GUI screen driver.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Refactored to implement new GenericScreenDriver APIs and the support provided 
**                  by it.  Added mouse support and other misc APIs related to window management.
** 004 MAG 20140902 Integrate COLOR-TABLE. Added support for BGCOLOR and FGCOLOR color conversion.
**                  Added system colors used solely for widget drawing.  
** 005 MAG 20140912 Implement GUI status bar.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Added support to allow drawing of nested containers (the real location of the
**                  translated origin is required).  Misc changes related to image drawing and
**                  mouse processing.
** 009 VIG 20141022 TC.screenSize replaced with window.size.
** 010 HC  20141031 Implemented support for drawing highlighted areas.
** 011 MAG 20141029 Added clip rectangle parameter to draw method.
** 012 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
**                  The GUI driver ensures that only one instance is used, for each client.
** 013 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 014 EVL 20141116 Adding round rectangles draw and fill.
** 015 CA  20141123 Added APIs to allow enforcement of window stacking at OS level.
** 016 CA  20141125 Invalidate the current simulator, if the currently selected window is being 
**                  destroyed.  Added API to conditionally enable/disable OS-level events.
** 017 EVL 20141125 Adding dotted line draw capability.
** 018 CA  20141205 Added API to draw relative to window's real origin (top-left corner), 
**                  regardless of the current translated origin.
**                  Added API to register editable widgets (thus the driver can adjust the mouse 
**                  cursor accordingly).
** 019 CA  20141220 Added APIs for clipboard access.  Added APIs to allow text drawing and font 
**                  selection matching legacy sizes.
** 020 HC  20141207 Implemented runtime support for several SESSION attributes.
** 021 EVL 20141211 Adding tooltip worker registration and deregistration.
** 022 EVL 20150203 Adding image offset drawing capability.
** 023 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 024 CA  20150317 Added API to determine the font's max width.
** 025 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 026 HC  20150323 Removed business logic from config classes.
** 027 GES 20150505 Massively refactored this class to minimize the Swing-specific logic and move
**                  the maximum logic to a new generic abstract base class.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.terminal.*;

/**
 * Swing-backed GUI driver.
 */
public class SwingGuiDriver
extends AbstractGuiDriver<Font>
{
   /**
    * The key reader.  This is global per the entire application, no matter how many frames are
    * alive.
    */
   private static final ContextLocal<SwingKeyboardReader> keyReader = 
   new ContextLocal<SwingKeyboardReader>()
   {
      protected SwingKeyboardReader initialValue()
      {
         return new WinKeyboardReader();
      };
   };

   /** A special map containing the attributes of a underlined font. */
   private static Map<TextAttribute, Integer> UNDERLINE_FONT = new HashMap<>();

   static 
   {
      UNDERLINE_FONT.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
      UNDERLINE_FONT = Collections.unmodifiableMap(UNDERLINE_FONT);
   }
   
   /**
    * Create a new GUI screen driver.
    */
   public SwingGuiDriver()
   {
      super(new GuiPrimitivesImpl()
      {
         @Override
         public void sync()
         {
            SwingUtilities.invokeLater(new Runnable()
            {
               @Override
               public void run()
               {
                  ews.repaint();
               }
            });
         }
      });
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    */
   @Override
   public void beep()
   {
      Toolkit.getDefaultToolkit().beep();
   }
   
   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   @Override
   public int getScreenResolution()
   {
      return Toolkit.getDefaultToolkit().getScreenResolution();
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * <p>
    * The list of codes and detailed description is provided in {@link Keyboard}
    * class. The {@link KeyProcessor} can serve as an example of the key 
    * translation implementation.
    * 
    * @return  key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()                      
   {
      return keyReader.get().read();
   }

   /**
    * Create a new window emulator.
    * 
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newEmulator()
   {
      final EmulatedWindowState ews;

      try
      {
         ews = SwingGuiClient.load(keyReader.get());
         
         // the real window size is unknown at this time; it will be set later, when the layout
         // is performed
         ews.setScreenBitmap(new ScreenBitmap(0, 0));
      }
      
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new IllegalStateException("Could not create window emulator!", e);
      }
      
      return ews;
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   @Override
   public FontDetails<Font> createFont(String  name,
                                       int     size,
                                       boolean bold,
                                       boolean italic,
                                       boolean underline,
                                       byte[]  definition,
                                       int     legacyWidth,
                                       int     legacyHeight)
   {
      int style = 0;
      
      style += (bold ? Font.BOLD : 0);
      style += (italic ? Font.ITALIC : 0);
      
      // TODO: bitmap fonts
      
      Font f = new Font(name, style, size);
      
      if (underline)
      {
         f = f.deriveFont(UNDERLINE_FONT);
      }
      
      FontDetails<Font> fd = new FontDetails<Font>(f, legacyWidth, legacyHeight, definition);
      
      // TODO: if a new font is created from the byte array, then register it with the graphics
      // for all windows

      scaleFont(fd, legacyWidth, legacyHeight);
      
      return fd;
   }
   
   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    * @param    factor
    *           Scaling factor.
    */
   @Override
   public void deriveFont(FontDetails<Font> fd, float factor)
   {
      fd.font = fd.font.deriveFont(factor);
   }

   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToCliboard(String txt)
   {
      StringSelection stringSelection = new StringSelection(txt);
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      clipboard.setContents(stringSelection, stringSelection);
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   The clipboard contents.
    */
   @Override
   public String clipboardContents()
   {
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

      Transferable content = clipboard.getContents(null);
      boolean hasTransferableText = 
         (content != null && content.isDataFlavorSupported(DataFlavor.stringFlavor));
      
      String result = null;
      if (hasTransferableText)
      {
         try
         {
           result = (String) content.getTransferData(DataFlavor.stringFlavor);
         }
         catch (UnsupportedFlavorException | IOException ex)
         {
            // ignore
         }
       }
      
      return result;
   }
}
