/*
** Module   : GuiWebDriver.java
** Abstract : WebSocket-based GUI driver
**
** Copyright (c) 2013-2015 Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150325 Created initial version patterned off of the ChuiWebDriver.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

// TODO: remove this, it is there to allow compilation
import java.awt.Font;

// don't import the entire logging package as it has naming conflicts with jetty
import java.util.logging.Level;
import java.util.logging.Logger;

import java.io.*;

import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.util.resource.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;

/**
 * GUI driver using WebSockets to communicate with a very thin javascript client hosted in
 * a browser. This class creates and uses an embedded web server to accomplish this.
 */
public class GuiWebDriver
extends AbstractGuiDriver<Font>
implements EmbeddedWebServer,
           ClientProtocolHooks
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(GuiWebDriver.class.getName());

   /** Main page root target. */
   private static final String MAIN_TARGET = "/";

   /** Welcome page. */
   private static final String WELCOME_PAGE = "/index.html";
   
   /** Configuration. */
   private BootstrapConfig config;
   
   /** Delgated implementation for the embedded web server. */
   private EmbeddedWebServerImpl websrv;
   
   /** Web-Socket which provides a GUI-specific protocol for communication with the browser. */
   private GuiWebSocket websock;
   
   /** Lock object for shared syncrhonization of the web socket. */
   private Object lock = new Object();
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public GuiWebDriver(BootstrapConfig config)
   {
      super(new GuiPrimitivesImpl()
      {
         @Override
         public void sync()
         {
            ews.repaint();
         };
      });
      
      this.config = config;
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      long timeout = -1;
      long wdtimeout = -1;
      
      if (config != null)
      {
         timeout   = config.getInt("client", "web", "socketTimeout", -1);
         wdtimeout = config.getInt("client", "web", "watchdogTimeout", -1);
      }           

      // calculate our resource base
      Resource cpres = EmbeddedWebServerImpl.calcResourceBase(GuiWebDriver.class, "res");
     
      // calculate our welcome page
      String welcomeUrl = getWelcomePageUrl(MAIN_TARGET, WELCOME_PAGE);
      
      // create our handlers
      GuiWebPageHandler[] hdlrs = new GuiWebPageHandler[2]; 
      
      hdlrs[0] = new GuiWebPageHandler(config, MAIN_TARGET, welcomeUrl);         
      hdlrs[1] = new GuiWebPageHandler(config, WELCOME_PAGE, welcomeUrl);
      
      // create our web socket
      websock = new GuiWebSocket(lock, this, timeout, wdtimeout);
      
      // create the embedded server instance
      websrv = new EmbeddedWebServerImpl("GuiWebDriver",
                                         new GenericWebSocketCreator<WebClientProtocol>(websock),
                                         cpres,
                                         hdlrs);
   }

   /**
    * Called each time a connect occurs, this allows caller-controlled initialization of the
    * remote client.  There can only be 1 connection open at a time, but it is possible for
    * the connection to be opened, closed and then re-opened.
    * <p>
    * Whenever the main page is loaded or re-loaded, the WebSocket will connect and we must
    * put the virtual terminal into a known state.
    */
   @Override
   public void initRemoteClient()
   {
   }
   
   /**
    * Input key events when in VT100 mode.
    * 
    * @param    keyCode
    *           The key code from the input event.
    * @param    charCode
    *           The character code from the input event.
    */
   @Override
   public void injectVT100Key(int keyCode, char charCode)
   {
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    */
   @Override
   public void beep()
   {
      if (websock != null)
      {
         websock.beep();
      }
   }

   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return   The pressed key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()                      
   {
      return websock.readKey();
   }

   /**
    * Create a new window emulator.
    * 
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newEmulator()
   {
      final GuiWebEmulatedWindow ews;

      // try
      // {
         ews = new GuiWebEmulatedWindow(lock);
         
         // the real window size is unknown at this time; it will be set later, when the layout
         // is performed
         ews.setScreenBitmap(new ScreenBitmap(0, 0));
      // }
      // 
      // catch (InterruptedException e)
      // {
      //    throw new IllegalStateException("Could not create window emulator!", e);
      // }
      
      return ews;
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   @Override
   public FontDetails<Font> createFont(String  name,
                                       int     size,
                                       boolean bold,
                                       boolean italic,
                                       boolean underline,
                                       byte[]  definition,
                                       int     legacyWidth,
                                       int     legacyHeight)
   {
      // TODO
      return null;
   }
   
   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    * @param    factor
    *           Scaling factor.
    */
   @Override
   public void deriveFont(FontDetails<Font> fd, float factor)
   {
      // TODO
   }

   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   @Override
   public int getScreenResolution()
   {
      // TODO
      return 0;
   }
   
   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToCliboard(String txt)
   {
      // TODO
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   The clipboard contents.
    */
   @Override
   public String clipboardContents()
   {
      // TODO
      return "";
   }
   
   /**
    * Perform driver shutdown.
    */
   @Override
   public void shutdown()
   {
      // close the emulated windows first
      super.shutdown();
      
      // notify the remote client about application exit    
      websock.quit();
      
      // since the application will exit shutdown the server
      shutdownServer();
   }
   
   /**
    * Startup the embedded server using the server's key store to build the web server SSL
    * context.
    * <p>
    * There are two page handlers.  One is for the root ("/") target and the other is for
    * the index ("/index.html") target.  Both load the same page index.html.
    * <p>
    * When the user is first redirected the URL is <code>https://server:port/</code> and the
    * target is "/". The index.html page is loaded for this target.
    * <p>
    * After redirection and the main page is loaded, the URL in the browser address bar will
    * become <code>https://server:port/index.html</code> and the corresponding target is now
    * "/index.html".
    * <p>
    * When the user reloads the main page within the browser environment the "/index.html"
    * target is sent to the web server and the same main page index.html is loaded. 
    * 
    * @param    keyStore
    *           The exported server KeyStore. 
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      return websrv.startupServer(keyStore, config, host, port);
   }
   
   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      websrv.shutdownServer();
   }
   
   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      return websrv.getServerURI();
   }
   
   /**
    * Build the name of the main page as a resource locator string.
    * 
    * @param   pageContext
    *          Page context root.
    * @param   pageFile
    *          Main page HTML file.
    * 
    * @return  Resource name for the main page.
    */
   private String getWelcomePageUrl(String pageContext, String pageFile)
   {
      StringBuilder sb = new StringBuilder(pageContext);
      String packageName = GuiWebPageHandler.class.getPackage().getName();
      return sb.append(packageName.replace('.', '/')).append(pageFile).toString();
   }   
}
