/*
** Module   : InteractiveChildProcess.java
** Abstract : API for starting interactive child processes
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150409 First version. This interface helps isolate process launching from UI.
*/

package com.goldencode.p2j.util;

/**
 * Provides an API to launch and manage an interactive child process which is independent from
 * any specific UI modality. 
 */
public interface InteractiveChildProcess
{
   /**
    * Reports if the client is in batch mode.
    *
    * @return   <code>true</code> if in batch mode.
    */
   public boolean inBatchMode();
   
   /**
    * Reports if the client is missing an already initialized native terminal session.
    *
    * @return   <code>true</code> if this is an artificial terminal (there is no native
    *           terminal).
    */
   public boolean isArtificialTerm();
   
   /**
    * Launch the specified command line as an interactive child process.
    *
    * @param    cmdline
    *           The command line to execute.  The first element must be the program name.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    *
    * @return   The operating system process ID (pid) of the child process.
    */
   public int launch(String[] cmdline, boolean silent);
   
   /**
    * Block execution until the interactive child process has exited. This will ensure that
    * any operating system resources are released before returning.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    */
   public void waitForExit(int pid, boolean silent)
   throws InterruptedException;
   
   /**
    * Asynchronously cleanup any operating system resources when the interactive child process
    * has exited.  This processing should be done an a secondary thread and this method will
    * return immediately after starting that thread.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    */
   public void asynchCleanup(int pid);
   
   /**
    * Wait for the user to press a key.
    *
    * @return   The key code of the key pressed to clear the pause or -1 on any failure.
    */
   public int pauseForUser();

   /**
    * Temporarily halts the client UI processing (key reading, normal drawing
    * mode)  After entering suspend mode, the current contents of the
    * terminal are cleared.  Use {@link #resume} to return to normal mode.
    */
   public void suspend();
   
   /**
    * Re-enables client UI processing (key reading, normal drawing mode).
    * Before resuming the current contents of the terminal are cleared.  This
    * is the mechanism to return to normal processing mode after having
    * called {@link #suspend}.
    */
   public void resume();   
}
